/* vim: syn=cpp
 *
 * SNAFU - Letting go of daily frustration
 *
 * Copyright (C) 2016  B.Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <util/atomic.h>
#include <avr/interrupt.h>
#include <avr/cpufunc.h>
#include <Wire.h>

#include "RTClib.h"
#include "snafucode.h"
#include "notes.h"
#include "serial.h"
#include "timing.h"

/*
 * Bit positions withing the shift registers for the different words.
 */
#define LPOS(x,y)	(((x) << 4) | (y))
#define WPOS(x)		(((x) >> 4) & 0x0f)
#define BPOS(x)		((x) & 0x0f)

#define WIX_ZERO	LPOS(0, 7)
#define WIX_ONE		LPOS(0, 6)
#define WIX_TWO		LPOS(3, 7)
#define WIX_THREE	LPOS(6, 7)
#define WIX_FOUR	LPOS(0, 5)
#define WIX_FIVE	LPOS(3, 6)
#define WIX_SIX		LPOS(3, 5)
#define WIX_SEVEN	LPOS(6, 6)
#define WIX_EIGHT	LPOS(0, 4)
#define WIX_NINE	LPOS(3, 3)
#define WIX_TEN		LPOS(3, 4)
#define WIX_TWENTY	LPOS(6, 5)
#define WIX_THIRTY	LPOS(0, 3)
#define WIX_FORTY	LPOS(3, 2)
#define WIX_FIFTY	LPOS(3, 1)
#define WIX_MINUTES	LPOS(0, 2)
#define WIX_HOURS	LPOS(0, 1)
#define WIX_DAYS	LPOS(6, 4)
#define WIX_WEEKS	LPOS(6, 3)
#define WIX_MONTHS	LPOS(0, 0)
#define WIX_LAST	LPOS(3, 0)
#define WIX_SNAFU	LPOS(6, 2)
#define WIX_ARRRGH	LPOS(6, 1)
#define WIX_BRAINFART	LPOS(6, 0)

/* Mask for updating words for partial colors */
#define MSK_R	1
#define MSK_G	2
#define MSK_B	4
#define MSK_RG	(MSK_R | MSK_G)
#define MSK_RB	(MSK_R | MSK_B)
#define MSK_GB	(MSK_G | MSK_B)
#define MSK_RGB	(MSK_R | MSK_G | MSK_B)

/* Word enumeration for logical setting of words */
enum {
	VAL_ZERO	= 0,
	VAL_ONE,
	VAL_TWO,
	VAL_THREE,
	VAL_FOUR,
	VAL_FIVE,
	VAL_SIX,
	VAL_SEVEN,
	VAL_EIGHT,
	VAL_NINE,
	VAL_TEN,
	VAL_TWENTY	= 20,
	VAL_THIRTY	= 30,
	VAL_FORTY	= 40,
	VAL_FIFTY	= 50,
	VAL_MINUTES	= 60,
	VAL_HOURS,
	VAL_DAYS,
	VAL_WEEKS,
	VAL_MONTHS,
	VAL_LAST,
	VAL_SNAFU,
	VAL_ARRRGH,
	VAL_BRAINFART,
	VAL_MAX,
};

/*
 * All word IDs
 */
static const uint8_t ledfields[] PROGMEM = {
	VAL_ZERO, VAL_ONE, VAL_TWO, VAL_THREE,
	VAL_FOUR, VAL_FIVE, VAL_SIX, VAL_SEVEN,
	VAL_EIGHT, VAL_NINE, VAL_TEN, VAL_TWENTY,
	VAL_THIRTY, VAL_FORTY, VAL_FIFTY, VAL_MINUTES,
	VAL_HOURS, VAL_DAYS, VAL_WEEKS, VAL_MONTHS,
	VAL_LAST, VAL_SNAFU, VAL_ARRRGH, VAL_BRAINFART,
};

// Events from buttons and timers
#define EV_BUT_SN_PRESS		0x01
#define EV_BUT_AR_PRESS		0x02
#define EV_BUT_BF_PRESS		0x04
#define EV_TIMER		0x10
#define EV_TIMER_SEC		0x20
#define EV_TIMEOUT		0x40
#define EV_TONETIMER		0x80

#define TIME_DEBOUNCE		8
#define TIME_LOCKOUT		15000		// Prevent fast button-press
#define TIME_SECOND		1000
#define TIME_FADEIN		100
#define TIME_FADEWAIT		1500
#define TIME_FADEOUT		125
#define TIME_SHOWPRESS		10000
#define TIME_BLINKTIME		200
#define TIME_DISPSEQ		1000
#define TIME_DISPRANDOM		100		// Interval between random colors
#define TIME_DISPRAINBOW	50
#define TIME_DISPSNARBF		50
#define TIME_DISPLEDSEQ		100

#define N_DISPRANDOM	51
#define N_DISPRAINBOW	85
#define N_DISPSNARBF	160

/* Pin definitions */
// 0 -> Serial RX
// 1 -> Serial TX
#define PIN_AUX_D2	2
#define PIN_AUX_D3	3
#define PIN_AUX_D4	4
#define PIN_AUX_D5	5
#define PIN_DEBUG	PIN_AUX_D2
#define PIN_BUT_SN	6
#define PIN_BUT_AR	7
#define PIN_BUT_BF	8
#define PIN_SOUND	9		// Buzzer output (PB1/OC1A)
#define PIN_VOLUME	A0		// Volume setting
#define PIN_LATCH	10
#define PIN_MOSI	11
#define PIN_MISO	12
#define PIN_SCK		13
// A4 -> I2C SDA
// A5 -> I2C SCL

#define dummyread(port)	do { asm volatile("in __tmp_reg__, %0" "\n\t" :  : "I" (_SFR_IO_ADDR(port))); } while(0)

// States of the machine
enum {
	ST_INIT,
	ST_DISP_SN,
	ST_DISP_AR,
	ST_DISP_BF,
	ST_DISP_BLINK,
	ST_DISPSEQ,
	ST_DISPSEQ_XX,
	ST_DISPSEQ_SN,
	ST_DISPSEQ_AR,
	ST_DISPSEQ_BF,
	ST_DISP_RANDOM,
	ST_DISP_RAINBOW,
	ST_DISP_SNARBF,
	ST_DISP_LEDSEQ,
};

#define NBBUF	12
static backbuf_t backbuf[NBBUF];	// 595 shift register's backing buffer(s)

static uint8_t bbcnt;		// Backbuf byte counter for shift-out
static uint8_t *bbptr;		// Backbuf pointer to shift-out data

static uint32_t oldtime;	// Previous' loop millis() value

static uint8_t bstate_sn;	// Button states
static uint8_t bstate_ar;
static uint8_t bstate_bf;
static uint8_t debtmr_sn;	// Button debounce timers
static uint8_t debtmr_ar;
static uint8_t debtmr_bf;

static uint16_t tmr_lo_sn;	// Lockout timers to prevent fast sequential pushes
static uint16_t tmr_lo_ar;
static uint16_t tmr_lo_bf;

static uint16_t timer;		// General timer
static uint16_t timerto;	// Timeout timer
static uint16_t timersec;	// Second timer
static uint16_t tonetimer;	// Tone duration timer

static const note_t *tunep;	// Current tune playing

timing_t time_xx[3];		// Time since last...

static uint8_t thestate;	// Current statemachine state
static uint8_t ndispseq;	// Sequence counter for display sequences

RTC_DS1307 rtc;			// Our clock chip
DateTime currenttime;		// Where we are now (read from rtc)
uint32_t currentstamp;		// in Unix epoch

/*
 * Local very fast Pseudo Random Number Generator (PRNG)
 * based on a simple LFSR design.
 */
static volatile uint32_t prng_val = 0x31415926;	// Local random number generator state
uint8_t lfsr_prng(void)
{
	uint8_t c = prng_val & 1;
	prng_val >>= 1;
	if(c)
		prng_val ^= 0xa6a6a6a6;
	return prng_val ^ (prng_val >> 8);
}

void lfsr_prng_mix(uint16_t v)
{
	prng_val ^= (uint32_t)v << 16;
}

/*
 * Set the speaker to generate a tone
 */
static inline void tone_set(uint8_t note, uint16_t duration)
{
	tonetimer = duration;
	if(note && note <= NTONE_PITCH) {
		OCR1A = pgm_read_word(&tone_pitch[note-1]);
		TCCR1A = _BV(COM1A0);	/* Set toggle output */
		TCNT1 = 0;
	} else
		sound_off();
}

/*
 * Play the next note of the tune
 */
static void tune_next()
{
	if(tunep) {
		uint8_t n = pgm_read_byte(&tunep->note);
		uint16_t d = pgm_read_word(&tunep->duration);
		if(n || d) {
			tone_set(n, d);
			tunep++;
		} else {
			tunep = NULL;
			tone_set(0, 0);
		}
	} else {
		tone_set(0, 0);
	}
}

/*
 * Volume management
 */
static inline void tune_volume(uint8_t hilo)
{
	digitalWrite(PIN_VOLUME, hilo);
}

/*
 * Play a specific tune
 */
static inline void tune_set(uint8_t tune, uint8_t vol)
{
	switch(tune) {
	case 0:	tunep = tune1; break;
	case 1:	tunep = tune2; break;
	case 2:	tunep = tune3; break;
	case 3:	tunep = tune4; break;
	case 4:	tunep = tune5; break;
	default:
		sound_off();
		return;
	}
	tune_next();
	tune_volume(vol);
}

/*
 * Debounce a button and return an event at the rising edge of the detection.
 * The rising edge ensures that there is no delay from pressing the button and
 * the event propagating. It is a prerequisite that the input line is not
 * glitchy.
 * A release event may be generated if the routine is slightly modified.
 */
static inline uint8_t do_debounce(uint8_t tdiff, uint8_t *bstate, uint8_t *debtmr, uint8_t pin, uint8_t ev)
{
	if(0 == *debtmr) {
		uint8_t state = digitalRead(pin);
		if(state != *bstate) {
			*debtmr = TIME_DEBOUNCE;
			if(!(*bstate = state))
				return ev;	// Event on High-to-Low transition of input
			// else
			//  return release_event_value
		}
	} else {
		if(*debtmr >= tdiff)
			*debtmr -= tdiff;
		else
			*debtmr = 0;
	}
	return 0;
}

/*
 * Timer countdown and return an event on timer reaching zero.
 */
static inline uint8_t do_timer(uint8_t tdiff, uint16_t *tmr, uint8_t ev)
{
	if(0 != *tmr) {
		if(*tmr >= tdiff)
			*tmr -= tdiff;	// Timer countdown
		else
			*tmr = 0;
		// Set event when done counting
		if(0 == *tmr)
			return ev;
	}
	return 0;
}

/*
 * Emit the latch signal to update the shift-register's output
 */
static inline void latch()
{
	digitalWrite(PIN_LATCH, 1);
	_NOP();
	_NOP();
	_NOP();
	_NOP();
	_NOP();
	_NOP();
	_NOP();
	_NOP();
	digitalWrite(PIN_LATCH, 0);
}

/*
 * SPI Interrupt
 * Output the next data to the 595 shift register
 */
ISR(SPI_STC_vect)
{
	// Entering here clears the SPIF
	// That is good because it won't get set after the last transfer

	if(bbcnt >= sizeof(backbuf[0])) {
		// The current colors are out, latch it and setup for next timer kick
		latch();
		bbcnt = 0;
		if(bbptr >= (uint8_t *)&backbuf[NBBUF])
			bbptr = (uint8_t *)&backbuf[0];
	} else {
		SPDR = *bbptr++;
		bbcnt++;
	}
}

/*
 * Timer 2 CTC Interrupt
 * Initiate the SPI interrupt train for output
 */
ISR(TIMER2_COMPA_vect)
{
	// The SPI buffer write will initiate a SPI transfer and set the
	// interrupt train in motion
	SPDR = *bbptr++;
	bbcnt++;
}

/*
 * Set the LED matrix
 * This is the low-level buffer data assignment for the LEDs
 */
static void matrix_led_set(mledset_t set, uint8_t row, uint8_t rowval, uint8_t hilo)
{
	rowval &= 0x0f;
	switch(set) {
	case MLED_SN:
		backbuf[row+0].m1 &= 0xf0;
		backbuf[row+4].m1 &= 0xf0;
		backbuf[row+8].m1 &= 0xf0;
		if(hilo--) {
			backbuf[row+0].m1 |= rowval;
			if(hilo--) {
				backbuf[row+4].m1 |= rowval;
				if(hilo) {
					backbuf[row+8].m1 |= rowval;
				}
			}
		}
		break;
	case MLED_AR:
		backbuf[row+0].m2 &= 0xf0;
		backbuf[row+4].m2 &= 0xf0;
		backbuf[row+8].m2 &= 0xf0;
		if(hilo--) {
			backbuf[row+0].m2 |= rowval;
			if(hilo--) {
				backbuf[row+4].m2 |= rowval;
				if(hilo) {
					backbuf[row+8].m2 |= rowval;
				}
			}
		}
		break;
	case MLED_BF:
		backbuf[row+0].m3 &= 0xf0;
		backbuf[row+4].m3 &= 0xf0;
		backbuf[row+8].m3 &= 0xf0;
		if(hilo--) {
			backbuf[row+0].m3 |= rowval & 0x0f;
			if(hilo--) {
				backbuf[row+4].m3 |= rowval;
				if(hilo) {
					backbuf[row+8].m3 |= rowval;
				}
			}
		}
		break;
	}
}

#ifdef COUNT_DEC
#define matrix_set	matrix_set_dec
#else
#define matrix_set	matrix_set_hex
#endif

#ifdef COUNT_DEC
/*
 * Show a decimal (BCD) value on the LED matrix
 */
static void matrix_set_dec(mledset_t set, uint16_t val, uint8_t hilo)
{
	uint8_t c1, c2, c3, c4;
	uint8_t r0 = 0, r1 = 0, r2 = 0, r3 = 0;
	// Split the value into decimal digits
	c1 = val % 10;
	val /= 10;
	c2 = val % 10;
	val /= 10;
	c3 = val % 10;
	val /= 10;
	c4 = val % 10;

	// Row/column swap right-bottom is 0,0
	if(c1 & 1)	r3 |= 1;
	if(c1 & 2)	r2 |= 1;
	if(c1 & 4)	r1 |= 1;
	if(c1 & 8)	r0 |= 1;
	if(c2 & 1)	r3 |= 2;
	if(c2 & 2)	r2 |= 2;
	if(c2 & 4)	r1 |= 2;
	if(c2 & 8)	r0 |= 2;
	if(c3 & 1)	r3 |= 4;
	if(c3 & 2)	r2 |= 4;
	if(c3 & 4)	r1 |= 4;
	if(c3 & 8)	r0 |= 4;
	if(c4 & 1)	r3 |= 8;
	if(c4 & 2)	r2 |= 8;
	if(c4 & 4)	r1 |= 8;
	if(c4 & 8)	r0 |= 8;
	matrix_led_set(set, 0, r0, hilo);
	matrix_led_set(set, 1, r1, hilo);
	matrix_led_set(set, 2, r2, hilo);
	matrix_led_set(set, 3, r3, hilo);
}

#else

/*
 * Show a hexadecimal value on the LED matrix
 */
static void matrix_set_hex(mledset_t set, uint16_t val, uint8_t hilo)
{
	uint8_t r0 = 0, r1 = 0, r2 = 0, r3 = 0;

	// Row/column swap right-bottom is 0,0
	if(val & (1 <<  0))	r3 |= 1;
	if(val & (1 <<  1))	r2 |= 1;
	if(val & (1 <<  2))	r1 |= 1;
	if(val & (1 <<  3))	r0 |= 1;
	if(val & (1 <<  4))	r3 |= 2;
	if(val & (1 <<  5))	r2 |= 2;
	if(val & (1 <<  6))	r1 |= 2;
	if(val & (1 <<  7))	r0 |= 2;
	if(val & (1 <<  8))	r3 |= 4;
	if(val & (1 <<  9))	r2 |= 4;
	if(val & (1 << 10))	r1 |= 4;
	if(val & (1 << 11))	r0 |= 4;
	if(val & (1 << 12))	r3 |= 8;
	if(val & (1 << 13))	r2 |= 8;
	if(val & (1 << 14))	r1 |= 8;
	if(val & (1 << 15))	r0 |= 8;
	matrix_led_set(set, 0, r0, hilo);
	matrix_led_set(set, 1, r1, hilo);
	matrix_led_set(set, 2, r2, hilo);
	matrix_led_set(set, 3, r3, hilo);
}
#endif

/*
 * Show an arbitray pattern on the LED matrix
 */
static void matrix_set_img(mledset_t set, uint16_t val, uint8_t hilo)
{
	uint8_t r0 = (val >>  0) & 0x0f;
	uint8_t r1 = (val >>  4) & 0x0f;
	uint8_t r2 = (val >>  8) & 0x0f;
	uint8_t r3 = (val >> 12) & 0x0f;
	matrix_led_set(set, 0, r0, hilo);
	matrix_led_set(set, 1, r1, hilo);
	matrix_led_set(set, 2, r2, hilo);
	matrix_led_set(set, 3, r3, hilo);
}

#define NCLEVELS	8
#define MAXINTENSITY	7
// Distribute the on-time between the RGB channels to distribute the current
// drawn on PWM.
static const uint16_t intensitybitmap_r[NCLEVELS] PROGMEM = {
	0x000, 0x001, 0x011, 0x111, 0x131, 0x313, 0x737, 0xfff
};
static const uint16_t intensitybitmap_g[NCLEVELS] PROGMEM = {
	0x000, 0x010, 0x104, 0x222, 0x262, 0x626, 0xe6e, 0xfff
};
static const uint16_t intensitybitmap_b[NCLEVELS] PROGMEM = {
	0x000, 0x100, 0x410, 0x444, 0x4c4, 0xc4c, 0xcdd, 0xfff
};

// Saturated color map
// Circles the outside boundary of the HSV circle.
// Colors are not luminance compensated. This would require an interaction with
// the intensity map above, which is too much work for now.
static const uint16_t colors[] PROGMEM = {
	0x0007, 0x0017, 0x0027, 0x0037, 0x0047, 0x0057, 0x0067,
	0x0077, 0x0076, 0x0075, 0x0074, 0x0073, 0x0072, 0x0071,
	0x0070, 0x0170, 0x0270, 0x0370, 0x0470, 0x0570, 0x0670,
	0x0770, 0x0760, 0x0750, 0x0740, 0x0730, 0x0720, 0x0710,
	0x0700, 0x0701, 0x0702, 0x0703, 0x0704, 0x0705, 0x0706,
	0x0707, 0x0607, 0x0507, 0x0407, 0x0307, 0x0207, 0x0107,
};

static void words_led_set_r(uint8_t wrd, uint8_t r)
{
	uint16_t bits;
	if(r >= NELEM(intensitybitmap_r))
		r = NELEM(intensitybitmap_r)-1;
	bits = pgm_read_word(&intensitybitmap_r[r]);
	for(uint8_t i = 0; i < NBBUF; i++) {
		if(bits & 1)
			((uint8_t *)&backbuf[i])[WPOS(wrd)+0] |= _BV(BPOS(wrd));
		else
			((uint8_t *)&backbuf[i])[WPOS(wrd)+0] &= ~_BV(BPOS(wrd));
		bits >>= 1;
	}
}

static void words_led_set_g(uint8_t wrd, uint8_t g)
{
	uint16_t bits;
	if(g >= NELEM(intensitybitmap_g))
		g = NELEM(intensitybitmap_g)-1;
	bits = pgm_read_word(&intensitybitmap_g[g]);
	for(uint8_t i = 0; i < NBBUF; i++) {
		if(bits & 1)
			((uint8_t *)&backbuf[i])[WPOS(wrd)+1] |= _BV(BPOS(wrd));
		else
			((uint8_t *)&backbuf[i])[WPOS(wrd)+1] &= ~_BV(BPOS(wrd));
		bits >>= 1;
	}
}

static void words_led_set_b(uint8_t wrd, uint8_t b)
{
	uint16_t bits;
	if(b >= NELEM(intensitybitmap_b))
		b = NELEM(intensitybitmap_b)-1;
	bits = pgm_read_word(&intensitybitmap_b[b]);
	for(uint8_t i = 0; i < NBBUF; i++) {
		if(bits & 1)
			((uint8_t *)&backbuf[i])[WPOS(wrd)+2] |= _BV(BPOS(wrd));
		else
			((uint8_t *)&backbuf[i])[WPOS(wrd)+2] &= ~_BV(BPOS(wrd));
		bits >>= 1;
	}
}

static inline void words_led_set(uint8_t wrd, uint8_t r, uint8_t g, uint8_t b, uint8_t msk)
{
	if(msk & 1)	words_led_set_r(wrd, r);
	if(msk & 2)	words_led_set_g(wrd, g);
	if(msk & 4)	words_led_set_b(wrd, b);
}

/*
 * Clear the first 9 bytes of each backbuf part
 * These are the RGB values of the words
 */
static void words_clear(void)
{
	for(uint8_t i = 0; i < NELEM(backbuf); i++) {
		uint8_t * ptr = (uint8_t *)&backbuf[i];
		for(uint8_t j = 0; j < 9; j++)
			*ptr++ = 0;
	}
}

/*
 * Set the decimal digit value of a word
 */
static void words_set_10(uint8_t val, uint8_t r, uint8_t g, uint8_t b, uint8_t msk)
{
	switch(val) {
	case 0:		words_led_set(WIX_ZERO, r, g, b, msk); break;
	case 1:		words_led_set(WIX_ONE, r, g, b, msk); break;
	case 2:		words_led_set(WIX_TWO, r, g, b, msk); break;
	case 3:		words_led_set(WIX_THREE, r, g, b, msk); break;
	case 4:		words_led_set(WIX_FOUR, r, g, b, msk); break;
	case 5:		words_led_set(WIX_FIVE, r, g, b, msk); break;
	case 6:		words_led_set(WIX_SIX, r, g, b, msk); break;
	case 7:		words_led_set(WIX_SEVEN, r, g, b, msk); break;
	case 8:		words_led_set(WIX_EIGHT, r, g, b, msk); break;
	case 9:		words_led_set(WIX_NINE, r, g, b, msk); break;
	case 10:	words_led_set(WIX_TEN, r, g, b, msk); break;
	case 20:	words_led_set(WIX_TWENTY, r, g, b, msk); break;
	case 30:	words_led_set(WIX_THIRTY, r, g, b, msk); break;
	case 40:	words_led_set(WIX_FORTY, r, g, b, msk); break;
	case 50:	words_led_set(WIX_FIFTY, r, g, b, msk); break;
	}
}

/*
 * Set the numerical value of a word
 * Or the extra words
 */
static void words_set(uint8_t val, uint8_t r, uint8_t g, uint8_t b, uint8_t msk)
{
	if(val < 60) {
		uint8_t u = val % 10;
		uint8_t t = val - u;
		if(u || !val)
			words_set_10(u, r, g, b, msk);
		if(t)
			words_set_10(t, r, g, b, msk);
	} else {
		switch(val) {
		case VAL_MINUTES:	words_led_set(WIX_MINUTES, r, g, b, msk); break;
		case VAL_HOURS:		words_led_set(WIX_HOURS, r, g, b, msk); break;
		case VAL_DAYS:		words_led_set(WIX_DAYS, r, g, b, msk); break;
		case VAL_WEEKS:		words_led_set(WIX_WEEKS, r, g, b, msk); break;
		case VAL_MONTHS:	words_led_set(WIX_MONTHS, r, g, b, msk); break;
		case VAL_LAST:		words_led_set(WIX_LAST, r, g, b, msk); break;
		case VAL_SNAFU:		words_led_set(WIX_SNAFU, r, g, b, msk); break;
		case VAL_ARRRGH:	words_led_set(WIX_ARRRGH, r, g, b, msk); break;
		case VAL_BRAINFART:	words_led_set(WIX_BRAINFART, r, g, b, msk); break;
		default:
			break;
		}
	}
}

static const struct __patterns_t {
	uint16_t	sn;
	uint16_t	ar;
	uint16_t	bf;
} patterns[] PROGMEM = {
	{ 0x000f, 0x0000, 0x0000 },
	{ 0x00f0, 0x0000, 0x0000 },
	{ 0x0f00, 0x0000, 0x0000 },
	{ 0xf000, 0x0000, 0x0000 },
	{ 0x0000, 0x000f, 0x0000 },
	{ 0x0000, 0x00f0, 0x0000 },
	{ 0x0000, 0x0f00, 0x0000 },
	{ 0x0000, 0xf000, 0x0000 },
	{ 0x0000, 0x0000, 0x000f },
	{ 0x0000, 0x0000, 0x00f0 },
	{ 0x0000, 0x0000, 0x0f00 },
	{ 0x0000, 0x0000, 0xf000 },
	{ 0x0000, 0x0000, 0x0f00 },
	{ 0x0000, 0x0000, 0x00f0 },
	{ 0x0000, 0x0000, 0x000f },
	{ 0x0000, 0xf000, 0x0000 },
	{ 0x0000, 0x0f00, 0x0000 },
	{ 0x0000, 0x00f0, 0x0000 },
	{ 0x0000, 0x000f, 0x0000 },
	{ 0xf000, 0x0000, 0x0000 },
	{ 0x0f00, 0x0000, 0x0000 },
	{ 0x00f0, 0x0000, 0x0000 },
	{ 0x000f, 0x0000, 0x0000 },

	{ 0x0000, 0x0000, 0x0000 },

	{ 0x1111, 0x1111, 0x1111 },
	{ 0x2222, 0x2222, 0x2222 },
	{ 0x4444, 0x4444, 0x4444 },
	{ 0x8888, 0x8888, 0x8888 },
	{ 0x4444, 0x4444, 0x4444 },
	{ 0x2222, 0x2222, 0x2222 },
	{ 0x1111, 0x1111, 0x1111 },
	{ 0x2222, 0x2222, 0x2222 },
	{ 0x4444, 0x4444, 0x4444 },
	{ 0x8888, 0x8888, 0x8888 },
	{ 0x4444, 0x4444, 0x4444 },
	{ 0x2222, 0x2222, 0x2222 },
	{ 0x1111, 0x1111, 0x1111 },

	{ 0x0000, 0x0000, 0x0000 },

	{ 0x0001, 0x0000, 0x0000 },
	{ 0x0012, 0x0000, 0x0000 },
	{ 0x0124, 0x0000, 0x0000 },
	{ 0x1248, 0x0000, 0x0000 },
	{ 0x2480, 0x0001, 0x0000 },
	{ 0x4800, 0x0012, 0x0000 },
	{ 0x8000, 0x0124, 0x0000 },
	{ 0x0000, 0x1248, 0x0000 },
	{ 0x0000, 0x2480, 0x0001 },
	{ 0x0000, 0x4800, 0x0012 },
	{ 0x0000, 0x8000, 0x0124 },
	{ 0x0000, 0x0000, 0x1248 },
	{ 0x0000, 0x0000, 0x2480 },
	{ 0x0000, 0x0000, 0x4800 },
	{ 0x0000, 0x0000, 0x8000 },

	{ 0x0008, 0x0000, 0x0000 },
	{ 0x0084, 0x0000, 0x0000 },
	{ 0x0842, 0x0000, 0x0000 },
	{ 0x8421, 0x0000, 0x0000 },
	{ 0x4210, 0x0008, 0x0000 },
	{ 0x2100, 0x0084, 0x0000 },
	{ 0x1000, 0x0842, 0x0000 },
	{ 0x0000, 0x8421, 0x0000 },
	{ 0x0000, 0x4210, 0x0008 },
	{ 0x0000, 0x2100, 0x0084 },
	{ 0x0000, 0x1000, 0x0842 },
	{ 0x0000, 0x0000, 0x8421 },
	{ 0x0000, 0x0000, 0x4210 },
	{ 0x0000, 0x0000, 0x2100 },
	{ 0x0000, 0x0000, 0x1000 },

	{ 0x0000, 0x0000, 0x0000 },

	{ 0x0c30, 0x1248, 0x4422 },
	{ 0x8421, 0x03c0, 0x2244 },
	{ 0x4422, 0x0c30, 0x1248 },
	{ 0x2244, 0x8421, 0x03c0 },
	{ 0x1248, 0x4422, 0x0c30 },
	{ 0x03c0, 0x2244, 0x8421 },
	{ 0x0c30, 0x1248, 0x4422 },
	{ 0x8421, 0x03c0, 0x2244 },
	{ 0x4422, 0x0c30, 0x1248 },
	{ 0x2244, 0x8421, 0x03c0 },
	{ 0x1248, 0x4422, 0x0c30 },
	{ 0x03c0, 0x2244, 0x8421 },
	{ 0x0c30, 0x1248, 0x4422 },
	{ 0x8421, 0x03c0, 0x2244 },
	{ 0x4422, 0x0c30, 0x1248 },
	{ 0x2244, 0x8421, 0x03c0 },
	{ 0x1248, 0x4422, 0x0c30 },
	{ 0x03c0, 0x2244, 0x8421 },
	{ 0x0c30, 0x1248, 0x4422 },
	{ 0x8421, 0x03c0, 0x2244 },
	{ 0x4422, 0x0c30, 0x1248 },
	{ 0x2244, 0x8421, 0x03c0 },
	{ 0x1248, 0x4422, 0x0c30 },
	{ 0x03c0, 0x2244, 0x8421 },

	{ 0x0000, 0x0000, 0x0000 },
};

static const uint8_t ledflash[] PROGMEM = {
	0, 1, 2, 3, 2, 1, 0, 1, 2, 3, 2, 1, 0, 1, 2, 3, 2, 1, 0,
};

static uint8_t disp_ledseq(uint8_t *seq)
{
	if(*seq < NELEM(ledflash)) {
		uint8_t l = pgm_read_byte(&ledflash[*seq]);
		matrix_set_img(MLED_SN, 0xffff, l);
		matrix_set_img(MLED_AR, 0xffff, l);
		matrix_set_img(MLED_BF, 0xffff, l);
	} else if(*seq < NELEM(ledflash) + NELEM(patterns)) {
		matrix_set_img(MLED_SN, pgm_read_word(&patterns[*seq - NELEM(ledflash)].sn), 3);
		matrix_set_img(MLED_AR, pgm_read_word(&patterns[*seq - NELEM(ledflash)].ar), 3);
		matrix_set_img(MLED_BF, pgm_read_word(&patterns[*seq - NELEM(ledflash)].bf), 3);
	} else {
		*seq = (uint8_t)0;
		return 1;
	}
	(*seq)++;
	return 0;
}

static inline void pick_a_color(uint8_t *rr, uint8_t *gg, uint8_t *bb)
{
#if 1
	static const uint8_t primes[] PROGMEM = { 5, 7, 9, 11, 13 };
	static uint8_t ix;
	ix += pgm_read_byte(&primes[lfsr_prng() % NELEM(primes)]);
	ix %= NELEM(colors);
#else
	static uint8_t last;
	uint8_t ix = lfsr_prng() % NELEM(colors);
	uint8_t d = abs((int8_t)ix - (int8_t)last);
	if(d < 9)
		ix = last + 9;
	if(d > 21)
		ix = last + 21;
	ix %= NELEM(colors);
	last = ix;
#endif
	uint16_t cc = pgm_read_word(&colors[ix]);
	*rr = (cc >> 0) & 0x0f;
	*gg = (cc >> 4) & 0x0f;
	*bb = (cc >> 8) & 0x0f;
}

/*
 * Random colors on all words
 */
static void random_color_words(void)
{
	words_clear();
	for(uint8_t i = 0; i < NELEM(ledfields); i++) {
		words_set(pgm_read_byte(&ledfields[i]), lfsr_prng() & MAXINTENSITY, lfsr_prng() & MAXINTENSITY, lfsr_prng() & MAXINTENSITY, MSK_RGB);
	}
}

/*
 * Dislay time with words in sequence
 */
static inline void seq_set_value(const timing_t *t, uint8_t rr, uint8_t gg, uint8_t bb, uint8_t msk)
{
	if(t->months)
		words_set(t->months, rr, gg, bb, msk);
	else if(t->weeks && t->days >= 10)
		words_set(t->weeks, rr, gg, bb, msk);
	else if(t->days)
		words_set(t->days, rr, gg, bb, msk);
	else if(t->hours)
		words_set(t->hours, rr, gg, bb, msk);
	else
		words_set(t->minutes, rr, gg, bb, msk);
}

static inline void seq_set_scale(const timing_t *t, uint8_t rr, uint8_t gg, uint8_t bb, uint8_t msk)
{
	if(t->months)
		words_set(VAL_MONTHS, rr, gg, bb, msk);
	else if(t->weeks && t->days >= 10)
		words_set(VAL_WEEKS, rr, gg, bb, msk);
	else if(t->days)
		words_set(VAL_DAYS, rr, gg, bb, msk);
	else if(t->hours)
		words_set(VAL_HOURS, rr, gg, bb, msk);
	else
		words_set(VAL_MINUTES, rr, gg, bb, msk);
}

static inline uint8_t weeks_less_than_10(const timing_t *t)
{
	return t->days && t->weeks < 10;
}

static inline uint8_t days_less_than_10(const timing_t *t)
{
	return t->days && t->days < 10;
}

static inline uint8_t hours_overlap_minutes(const timing_t *t)
{
	return (t->hours % 10) == (t->minutes % 10) || (t->hours / 10) == (t->minutes / 10);
}

// If months > 0, then weeks is also > 0
// If weeks > 0, then days is also > 0
// If days == 0, then we are within the first 24 hours
static void dispseq_show_lower(const timing_t *t, uint8_t r, uint8_t g, uint8_t b, uint8_t msk)
{
	if(!t->days && !t->hours && t->minutes) {
		// XX Minutes, Zero Hours
		words_set(t->minutes, r, g, b, msk);
		words_set(VAL_MINUTES, r, g, b, msk);
	} else if(!t->days && !hours_overlap_minutes(t)) {
		// XX Hours, YY Minutes
		words_set(t->hours, r, g, b, msk);
		words_set(VAL_HOURS, r, g, b, msk);
	} else if(days_less_than_10(t) && t->days != (t->hours % 10)) {
		// XX Days, YY Hours
		words_set(t->days, r, g, b, msk);
		words_set(VAL_DAYS, r, g, b, msk);
	} else if(weeks_less_than_10(t) && t->weeks != (t->days % 7)) {
		// XX Weeks, YY Days
		words_set(t->weeks, r, g, b, msk);
		words_set(VAL_WEEKS, r, g, b, msk);
	} else if(t->weeks >= 10 && t->months != (t->days % 30) && (t->months % 10) != (t->days % 10)) {
		// XX Months, YY Days
		words_set(t->months, r, g, b, msk);
		words_set(VAL_MONTHS, r, g, b, msk);
	} else
		seq_set_value(t, r, g, b, msk);
}

static void dispseq_show_upper(const timing_t *t, uint8_t r, uint8_t g, uint8_t b, uint8_t msk)
{
	if(!t->days && !t->hours && t->minutes) {
		// XX Minutes, Zero Hours
		words_set(0, r, g, b, msk);
		words_set(VAL_HOURS, r, g, b, msk);
	} else if(!t->days && !hours_overlap_minutes(t)) {
		// XX Hours, YY Minutes
		words_set(t->minutes, r, g, b, msk);
		words_set(VAL_MINUTES, r, g, b, msk);
	} else if(days_less_than_10(t) && t->days != (t->hours % 10)) {
		// XX Days, YY Hours
		words_set(t->hours, r, g, b, msk);
		words_set(VAL_HOURS, r, g, b, msk);
	} else if(weeks_less_than_10(t) && t->weeks != (t->days % 7)) {
		// XX Weeks, YY Days
		words_set(t->days % 7, r, g, b, msk);
		words_set(VAL_DAYS, r, g, b, msk);
	} else if(t->weeks >= 10 && t->months != (t->days % 30) && (t->months % 10) != (t->days % 10)) {
		// XX Months, YY Days
		words_set(t->days % 30, r, g, b, msk);
		words_set(VAL_DAYS, r, g, b, msk);
	} else
		seq_set_scale(t, r, g, b, msk);
}

static void dispseq_seq(const timing_t *t, uint8_t *seq, uint8_t tag, uint8_t next)
{
	uint8_t rr, gg, bb;
	static timing_t tcpy;	// Local copy to prevent race conditions (set at first seq)

	switch((*seq)++) {
	default:
		*seq = 1;
	case 0:
		tcpy = *t;
		words_clear();
		//break;
		(*seq)++;
	case 1:
		switch(tag) {
		case VAL_SNAFU:		matrix_set(MLED_SN, time_sn.tm.events, 2); break;
		case VAL_ARRRGH:	matrix_set(MLED_AR, time_ar.tm.events, 2); break;
		case VAL_BRAINFART:	matrix_set(MLED_BF, time_bf.tm.events, 2); break;
		}
		pick_a_color(&rr, &gg, &bb);
		dispseq_show_lower(&tcpy, rr, gg, bb, MSK_RGB);
		break;
	case 2:
		pick_a_color(&rr, &gg, &bb);
		dispseq_show_upper(&tcpy, rr, gg, bb, MSK_RGB);
		break;
	case 3:
		pick_a_color(&rr, &gg, &bb);
		words_set(VAL_LAST, rr, gg, bb, MSK_RGB);
		break;
	case 4:
		switch(tag) {
		case VAL_SNAFU:		matrix_set(MLED_SN, time_sn.tm.events, 3); break;
		case VAL_ARRRGH:	matrix_set(MLED_AR, time_ar.tm.events, 3); break;
		case VAL_BRAINFART:	matrix_set(MLED_BF, time_bf.tm.events, 3); break;
		}
		pick_a_color(&rr, &gg, &bb);
		words_set(tag, rr, gg, bb, MSK_RGB);
		break;
	case 5:
		*seq = 0;
		state_set(next);
		break;
	}
}

/*
 * Show a moving rainbow over the words
 */
static void rainbow_words(void)
{
	static uint8_t cidx;
	uint8_t ci = cidx++;
	cidx %= NELEM(colors);

	for(uint8_t i = 0; i <= 9; i++) {
		uint16_t cc = pgm_read_word(&colors[ci]);
		if(++ci >= NELEM(colors))
			ci = 0;
		uint8_t r = (cc >> 0) & 0x0f;
		uint8_t g = (cc >> 4) & 0x0f;
		uint8_t b = (cc >> 8) & 0x0f;
		switch(i) {
		case 0:
			words_set(VAL_ZERO, r, g, b, MSK_RGB);
			words_set(VAL_ONE, r, g, b, MSK_RGB);
			words_set(VAL_TWO, r, g, b, MSK_RGB);
			words_set(VAL_THREE, r, g, b, MSK_RGB);
			break;
		case 1:
			words_set(VAL_FOUR, r, g, b, MSK_RGB);
			words_set(VAL_FIVE, r, g, b, MSK_RGB);
			words_set(VAL_SIX, r, g, b, MSK_RGB);
			words_set(VAL_SEVEN, r, g, b, MSK_RGB);
			break;
		case 2:
			words_set(VAL_EIGHT, r, g, b, MSK_RGB);
			words_set(VAL_NINE, r, g, b, MSK_RGB);
			words_set(VAL_TEN, r, g, b, MSK_RGB);
			words_set(VAL_TWENTY, r, g, b, MSK_RGB);
			break;
		case 4:
			words_set(VAL_THIRTY, r, g, b, MSK_RGB);
			words_set(VAL_FORTY, r, g, b, MSK_RGB);
			words_set(VAL_FIFTY, r, g, b, MSK_RGB);
			words_set(VAL_DAYS, r, g, b, MSK_RGB);
			break;
		case 5:
			words_set(VAL_MINUTES, r, g, b, MSK_RGB);
			words_set(VAL_HOURS, r, g, b, MSK_RGB);
			words_set(VAL_WEEKS, r, g, b, MSK_RGB);
			break;
		case 6:
			words_set(VAL_MONTHS, r, g, b, MSK_RGB);
			words_set(VAL_LAST, r, g, b, MSK_RGB);
			break;
		case 7:
			words_set(VAL_SNAFU, r, g, b, MSK_RGB);
			break;
		case 8:
			words_set(VAL_ARRRGH, r, g, b, MSK_RGB);
			break;
		case 9:
			words_set(VAL_BRAINFART, r, g, b, MSK_RGB);
			break;
		}
	}
}

/*
 * Fade the SN/AR/BF in seperate color channels
 */
static const uint8_t triangle[] PROGMEM = { 1, 2, 3, 4, 5, 6, 7, 6, 5, 4, 3, 2 };

static void snarbf_words(void)
{
	static uint8_t ri = 0;
	static uint8_t gi = 3;
	static uint8_t bi = 6;
	uint8_t r = pgm_read_byte(&triangle[ri]);
	uint8_t g = pgm_read_byte(&triangle[gi]);
	uint8_t b = pgm_read_byte(&triangle[bi]);

	dispseq_show_lower(&time_sn, r, 0, 0, MSK_R);
	dispseq_show_upper(&time_sn, r, 0, 0, MSK_R);
	words_set(VAL_SNAFU, r, 0, 0, MSK_RGB);
	dispseq_show_lower(&time_ar, 0, g, 0, MSK_G);
	dispseq_show_upper(&time_ar, 0, g, 0, MSK_G);
	words_set(VAL_ARRRGH, 0, g, 0, MSK_RGB);
	dispseq_show_lower(&time_bf, 0, 0, b, MSK_B);
	dispseq_show_upper(&time_bf, 0, 0, b, MSK_B);
	words_set(VAL_BRAINFART, 0, 0, b, MSK_RGB);

	if(++ri >= NELEM(triangle))
		ri = 0;
	if(++gi >= NELEM(triangle))
		gi = 0;
	if(++bi >= NELEM(triangle))
		bi = 0;
}

static void state_set(uint8_t newstate)
{
	// State exit handling
	switch(thestate) {
	case ST_DISPSEQ_SN:
	case ST_DISPSEQ_AR:
	case ST_DISPSEQ_BF:
	case ST_DISP_LEDSEQ:
		matrix_set(MLED_SN, time_sn.tm.events, 1);
		matrix_set(MLED_AR, time_ar.tm.events, 1);
		matrix_set(MLED_BF, time_bf.tm.events, 1);
		words_clear();
		break;
	}

	// Switch to new state
	thestate = newstate;

	// State entry handling
	switch(thestate) {
	case ST_INIT:
		matrix_set(MLED_SN, time_sn.tm.events, 1);
		matrix_set(MLED_AR, time_ar.tm.events, 1);
		matrix_set(MLED_BF, time_bf.tm.events, 1);
		break;

	case ST_DISP_SN:
	case ST_DISP_AR:
	case ST_DISP_BF:
		words_clear();
		words_set(VAL_ZERO, MAXINTENSITY, MAXINTENSITY, MAXINTENSITY, MSK_RGB);
		words_set(VAL_MINUTES, MAXINTENSITY, MAXINTENSITY, MAXINTENSITY, MSK_RGB);
		words_set(VAL_LAST, MAXINTENSITY, MAXINTENSITY, MAXINTENSITY, MSK_RGB);
		break;

	case ST_DISP_BLINK:
		tune_set(lfsr_prng() % NTUNES, 1);
		timerto = TIME_SHOWPRESS;
		timer = TIME_BLINKTIME;
		break;

	case ST_DISPSEQ:
		words_clear();
		matrix_set(MLED_SN, time_sn.tm.events, 1);
		matrix_set(MLED_AR, time_ar.tm.events, 1);
		matrix_set(MLED_BF, time_bf.tm.events, 1);
		break;

	case ST_DISPSEQ_XX:
		words_clear();
		timer = TIME_DISPSEQ;
		break;

	case ST_DISPSEQ_SN:
	case ST_DISPSEQ_AR:
	case ST_DISPSEQ_BF:
		timer = TIME_DISPSEQ;
		break;

	case ST_DISP_RANDOM:
		ndispseq = N_DISPRANDOM;
		timer = TIME_SECOND;
		break;

	case ST_DISP_RAINBOW:
		ndispseq = N_DISPRAINBOW;
		timer = TIME_SECOND;
		break;

	case ST_DISP_SNARBF:
		ndispseq = N_DISPSNARBF;
		timer = TIME_SECOND;
		break;

	case ST_DISP_LEDSEQ:
		ndispseq = 0;
		words_clear();
		timer = TIME_SECOND;
		break;
	}
}

/*
 * System initialization
 */
void setup()
{
	Serial.begin(115200);

	digitalWrite(PIN_LATCH, 0);
	digitalWrite(PIN_SCK, 0);
	pinMode(PIN_LATCH, OUTPUT);
	pinMode(PIN_MOSI, OUTPUT);
	pinMode(PIN_MISO, INPUT_PULLUP);	// Not used SPI mode makes this an input
	pinMode(PIN_SCK, OUTPUT);

	pinMode(PIN_BUT_SN, INPUT_PULLUP);
	pinMode(PIN_BUT_AR, INPUT_PULLUP);
	pinMode(PIN_BUT_BF, INPUT_PULLUP);

	pinMode(PIN_AUX_D2, OUTPUT);
	pinMode(PIN_AUX_D3, OUTPUT);
	pinMode(PIN_AUX_D4, OUTPUT);
	pinMode(PIN_AUX_D5, OUTPUT);

	digitalWrite(PIN_SOUND, 0);
	digitalWrite(PIN_VOLUME, 0);
	pinMode(PIN_SOUND, OUTPUT);
	pinMode(PIN_VOLUME, OUTPUT);

	// Setup ISR backbuf pointer and count
	bbcnt = 0;
	bbptr = (uint8_t *)&backbuf[0];

	memset(backbuf, 0, sizeof(backbuf));
	// Set the column driver lines
	backbuf[0].m1 = backbuf[0].m2 =	backbuf[0].m3 = backbuf[4].m1 = backbuf[4].m2 =	backbuf[4].m3 = backbuf[8].m1  = backbuf[8].m2  = backbuf[8].m3  = 0x10;
	backbuf[1].m1 = backbuf[1].m2 =	backbuf[1].m3 = backbuf[5].m1 = backbuf[5].m2 =	backbuf[5].m3 = backbuf[9].m1  = backbuf[9].m2  = backbuf[9].m3  = 0x20;
	backbuf[2].m1 = backbuf[2].m2 =	backbuf[2].m3 = backbuf[6].m1 = backbuf[6].m2 =	backbuf[6].m3 = backbuf[10].m1 = backbuf[10].m2 = backbuf[10].m3 = 0x40;
	backbuf[3].m1 = backbuf[3].m2 =	backbuf[3].m3 = backbuf[7].m1 = backbuf[7].m2 =	backbuf[7].m3 = backbuf[11].m1 = backbuf[11].m2 = backbuf[11].m3 = 0x80;

	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		// Setup SPI at Fosc/2, master, mode 0
		//SPSR = _BV(SPI2X);
		SPSR = 0;
		SPCR = _BV(SPE) | _BV(MSTR) | _BV(SPR0);
		SPDR = 0;
		while(!(SPSR & _BV(SPIF)))
			;
		dummyread(SPDR);	// Clear possibly pending interrupt
		SPCR |= _BV(SPIE);

		// Setup Timer 2 in CTC @1200 Hz to kick the SPI train
		// should give a 100 Hz PWM on the colors
		TCCR2A = _BV(WGM21);
		TCCR2B = _BV(CS22) | _BV(CS20);	// 1:128
		OCR2A = F_CPU / (128UL*1200UL) - 1;
		TCNT2 = 0;
		TIMSK2 = _BV(OCIE2A);
	}

	// Setup sound hardware with Timer 1.
	TCCR1A = 0;
#ifdef LOWNOTES
	TCCR1B = _BV(WGM12) | _BV(CS11);	// 1:8 CTC
#else
	TCCR1B = _BV(WGM12) | _BV(CS10);	// 1:1 CTC
#endif
	OCR1A = NOTE_C4;	// Just a value
	TCNT1 = 0;

	oldtime = millis();

	rtc.begin();		// Our RTC

	uint8_t rr = rtc.isrunning();
	if(!rr)
		rtc.adjust(DateTime(DateTime(2016, 4, 18, 20, 45, 0)));
	Serial.print(F("RTC run: "));
	Serial.println(rr);

	currenttime = rtc.now();
	currentstamp = currenttime.unixtime();

	Serial.print(F("Time is now   : "));
	Serial.println(currentstamp);

	timersec = TIME_SECOND;	// Set seconds timer

	timing_readall();	// Restore time and event count

	state_set(ST_INIT);

	serial_status();
	serial_menu();

	tune_volume(0);
	tone_set(NOTE_A0, DUR_1_4);
}

/*
 * Set the lockout timers
 * The main button is locked out full time
 * The other buttons are locked out half time
 */
static inline void set_lockout(uint16_t *t1, uint16_t *t2, uint16_t *t3)
{
	*t1 = TIME_LOCKOUT;
	if(*t2 < TIME_LOCKOUT / 2)
		*t2 = TIME_LOCKOUT / 2;
	if(*t3 < TIME_LOCKOUT / 2)
		*t3 = TIME_LOCKOUT / 2;
}

/*
 * Main program loop
 */
#define chk_ev(x)	((x) & events)

void loop()
{
	uint32_t now;
	uint8_t tdiff = (now = millis()) - oldtime;
	uint8_t events = 0;
	static uint8_t seq;
	static uint8_t tidx;

	/* Handle buttons and timers on (just about) every millisecond */
	if(tdiff) {
		oldtime = now;
		events |= do_debounce(tdiff, &bstate_sn, &debtmr_sn, PIN_BUT_SN, EV_BUT_SN_PRESS);
		events |= do_debounce(tdiff, &bstate_ar, &debtmr_ar, PIN_BUT_AR, EV_BUT_AR_PRESS);
		events |= do_debounce(tdiff, &bstate_bf, &debtmr_bf, PIN_BUT_BF, EV_BUT_BF_PRESS);
		events |= do_timer(tdiff, &timer, EV_TIMER);
		events |= do_timer(tdiff, &timerto, EV_TIMEOUT);
		events |= do_timer(tdiff, &timersec, EV_TIMER_SEC);
		events |= do_timer(tdiff, &tonetimer, EV_TONETIMER);
		do_timer(tdiff, &tmr_lo_sn, 0);
		do_timer(tdiff, &tmr_lo_ar, 0);
		do_timer(tdiff, &tmr_lo_bf, 0);
	}

	if(Serial.available() > 0)
		serial_input();

	// Throw off random when a button is pressed
	if(chk_ev(EV_BUT_SN_PRESS | EV_BUT_AR_PRESS | EV_BUT_BF_PRESS))
		lfsr_prng_mix(micros());

	// Kill any button events while in lockout
	// Restart the timer to extend any running timeout (user keeps pressing)
	if(tmr_lo_sn) {
		if(chk_ev(EV_BUT_SN_PRESS))
			set_lockout(&tmr_lo_sn, &tmr_lo_ar, &tmr_lo_bf);
		events &= ~EV_BUT_SN_PRESS;
	}
	if(tmr_lo_ar) {
		if(chk_ev(EV_BUT_AR_PRESS))
			set_lockout(&tmr_lo_ar, &tmr_lo_bf, &tmr_lo_sn);
		events &= ~EV_BUT_AR_PRESS;
	}
	if(tmr_lo_bf) {
		if(chk_ev(EV_BUT_BF_PRESS))
			set_lockout(&tmr_lo_bf, &tmr_lo_sn, &tmr_lo_ar);
		events &= ~EV_BUT_BF_PRESS;
	}
	if(chk_ev(EV_BUT_SN_PRESS)) {
		set_lockout(&tmr_lo_sn, &tmr_lo_ar, &tmr_lo_bf);
	}
	if(chk_ev(EV_BUT_AR_PRESS)) {
		set_lockout(&tmr_lo_ar, &tmr_lo_bf, &tmr_lo_sn);
	}
	if(chk_ev(EV_BUT_BF_PRESS)) {
		set_lockout(&tmr_lo_bf, &tmr_lo_sn, &tmr_lo_ar);
	}

	// Every second, update the local time and display counts
	if(chk_ev(EV_TIMER_SEC)) {
		currenttime = rtc.now();
		currentstamp = currenttime.unixtime();
		timersec = TIME_SECOND;
		timing_update(IDX_SN);
		timing_update(IDX_AR);
		timing_update(IDX_BF);
	}

	// Run the music in the background
	if(chk_ev(EV_TONETIMER))
		tune_next();

	// Check for any button press
	// Only one press is honored at any time
	if(chk_ev(EV_BUT_SN_PRESS)) {
		if(time_sn.tm.events < MAXCOUNT)
			time_sn.tm.events++;
		timing_reset(IDX_SN);
		state_set(ST_DISP_SN);
	} else if(chk_ev(EV_BUT_AR_PRESS)) {
		if(time_ar.tm.events < MAXCOUNT)
			time_ar.tm.events++;
		timing_reset(IDX_AR);
		state_set(ST_DISP_AR);
	} else if(chk_ev(EV_BUT_BF_PRESS)) {
		if(time_bf.tm.events < MAXCOUNT)
			time_bf.tm.events++;
		timing_reset(IDX_BF);
		state_set(ST_DISP_BF);
	}


	switch(thestate) {
	default:
	case ST_INIT:
		state_set(ST_DISPSEQ);
		break;

	case ST_DISPSEQ:
		state_set(ST_DISPSEQ_XX);
		seq = 0;
		break;

	case ST_DISPSEQ_XX:
		if((lfsr_prng() >> 1) & 0x1f) {
			state_set(ST_DISPSEQ_SN);
		} else {
			tune_set(lfsr_prng() % NTUNES, 0);
			switch((lfsr_prng() >> 3) & 3) {
			case 0: state_set(ST_DISP_RANDOM); break;
			case 1: state_set(ST_DISP_RAINBOW); break;
			case 2: state_set(ST_DISP_SNARBF); break;
			case 3: state_set(ST_DISP_LEDSEQ); break;
			}
		}
		break;

	case ST_DISPSEQ_SN:
		if(chk_ev(EV_TIMER)) {
			timer = TIME_DISPSEQ;
			dispseq_seq(&time_sn, &seq, VAL_SNAFU, ST_DISPSEQ_AR);
		}
		break;

	case ST_DISPSEQ_AR:
		if(chk_ev(EV_TIMER)) {
			timer = TIME_DISPSEQ;
			dispseq_seq(&time_ar, &seq, VAL_ARRRGH, ST_DISPSEQ_BF);
		}
		break;

	case ST_DISPSEQ_BF:
		if(chk_ev(EV_TIMER)) {
			timer = TIME_DISPSEQ;
			dispseq_seq(&time_bf, &seq, VAL_BRAINFART, ST_DISPSEQ_XX);
		}
		break;

	case ST_DISP_RANDOM:
		if(chk_ev(EV_TIMER)) {
			timer = TIME_DISPRANDOM;
			random_color_words();
			if(!--ndispseq)
				state_set(ST_DISPSEQ);
		}
		break;

	case ST_DISP_RAINBOW:
		if(chk_ev(EV_TIMER)) {
			timer = TIME_DISPRAINBOW;
			rainbow_words();
			if(!--ndispseq)
				state_set(ST_DISPSEQ);
		}
		break;

	case ST_DISP_SNARBF:
		if(chk_ev(EV_TIMER)) {
			timer = TIME_DISPSNARBF;
			snarbf_words();
			if(!--ndispseq)
				state_set(ST_DISPSEQ);
		}
		break;

	case ST_DISP_LEDSEQ:
		if(chk_ev(EV_TIMER)) {
			timer = TIME_DISPLEDSEQ;
			if(disp_ledseq(&ndispseq))
				state_set(ST_DISPSEQ);
		}
		break;

	case ST_DISP_SN:
		words_set(VAL_SNAFU, MAXINTENSITY, MAXINTENSITY, MAXINTENSITY, MSK_RGB);
		matrix_set(MLED_SN, time_sn.tm.events, 3);
		tidx = IDX_SN;
		state_set(ST_DISP_BLINK);
		break;

	case ST_DISP_AR:
		words_set(VAL_ARRRGH, MAXINTENSITY, MAXINTENSITY, MAXINTENSITY, MSK_RGB);
		matrix_set(MLED_AR, time_ar.tm.events, 3);
		tidx = IDX_AR;
		state_set(ST_DISP_BLINK);
		break;

	case ST_DISP_BF:
		words_set(VAL_BRAINFART, MAXINTENSITY, MAXINTENSITY, MAXINTENSITY, MSK_RGB);
		matrix_set(MLED_BF, time_bf.tm.events, 3);
		tidx = IDX_BF;
		state_set(ST_DISP_BLINK);
		break;

	case ST_DISP_BLINK:
		if(chk_ev(EV_TIMEOUT)) {
			state_set(ST_DISPSEQ);
			seq = 3;
			goto setiton;
		} else if(chk_ev(EV_TIMER)) {
			timer = TIME_BLINKTIME;
			if(++seq > 3)
				seq = 1;
setiton:
			switch(tidx) {
			case IDX_SN:
				matrix_set(MLED_SN, time_sn.tm.events, seq);
				words_set(VAL_SNAFU, MAXINTENSITY-2*(seq-1), MAXINTENSITY-2*(seq-1), MAXINTENSITY-2*(seq-1), MSK_RGB);
				break;
			case IDX_AR:
				matrix_set(MLED_AR, time_ar.tm.events, seq);
				words_set(VAL_ARRRGH, MAXINTENSITY-2*(seq-1), MAXINTENSITY-2*(seq-1), MAXINTENSITY-2*(seq-1), MSK_RGB);
				break;
			case IDX_BF:
				matrix_set(MLED_BF, time_bf.tm.events, seq);
				words_set(VAL_BRAINFART, MAXINTENSITY-2*(seq-1), MAXINTENSITY-2*(seq-1), MAXINTENSITY-2*(seq-1), MSK_RGB);
				break;
			}
		}
		break;
	}
}
